'''
Biologically-Inspired Mass Extinction in Evolutionary Algorithms
Author: Kaelan Engholdt
Version: 05/26/2021

Holds all extinction parameter sets for testing.
Consult Sub-Section 6.6 of the 'README.md' for more information.

'''

# TODO user must copy this file and place it in the same directory as their evolutionary algorithm

# TODO user must properly adjust all of the parameters they desire to use in the permutations

import ext_params as ep


# the starting extinction parameter set to be used
ext_parameter_set = 1

# extinction percentages to test (list must have 3 elements)
testing_ext_percent = [0.5, 0.6, 0.7]

# extinction durations to test (list must have 3 elements)
testing_ext_duration = [1, 10, 20]

# interval extinction intervals to test (list must have 3 elements)
testing_i_ext_interval = [10, 15, 20]

# probabilistic extinction intervals to test (list must have 3 elements)
testing_p_ext_interval = [10, 15, 20]

# probabilistic extinction probability increments to test (list must have 3 elements)
testing_p_ext_prob_inc = [0.05, 0.1, 0.15]

# repopulation durations to test (list must have 3 elements)
testing_repop_duration = [1, 10, 20]

# basic extinction parameters (used across all sets)
ext_sort_order = None            # indicates how the population is sorted, best to worst (True) or worst to best (False)
ext_kills_least_fit = None       # indicates if extinction kills the least fit members (True) or random members (False)
ext_kills_constant = None        # indicates if the amount extinction kills is constant (True) or logarithmic (False)
ext_virulence = None             # indicates if extinction gets stronger (True) or weaker (False) as it progresses
ext_pop_dyn_strength = 0.00      # specifies how strong population dynamics are, a value close to '0.15' is recommended
ext_end_percent = 0.00           # percent of generations that will execute before the extinction operator is turned off
ext_safeguard = 0                # minimum number of generations between possible extinction events

# elitism parameters (used across all sets)
ext_elite_percent = 0.1          # percent of population to consider elite (guaranteed to survive extinction)

# anti-elitism parameters (used across all sets)
ext_anti_elite_percent = 0.00    # percent of population to consider anti-elite (will be killed if an extinction occurs)

# probabilistic extinction parameters (used across all sets)
p_ext_base_prob = 0.30           # base extinction probability (resets to base probability after extinction event)
p_ext_always_resets = None       # indicates if probability always resets upon any extinction or only for probabilistic

# repopulation parameters (used across all sets)
repop_elite_percent = 0.00       # percent of members that will be generated by the elite members during repopulation
repop_survivor_percent = 0.60    # percent of members that will be generated by the current members of the population

# newborn parameters (used across all sets)
newborn_expires = 10             # number of generations repopulated members are considered newborn members

# alternate parameters (used across all sets)
alt_tourn_size = 0               # how much to decrease tournament size by when extinction occurs
alt_xo_prob = 0.20               # how much to increase the probability of crossover occurring by when extinction occurs
alt_mut_prob = 0.30              # how much to increase the probability of mutation occurring by when extinction occurs
alt_mut_rate = 0.00              # how much to increase the mutation rate by when extinction occurs


def saved_params(saved_idx):
    '''
    Description
    ----------
    Holds all permutations of the extinction parameters defined above to create 324 extinction parameter sets.
    To enable testing using these extinction parameter sets, enable 'ext_testing_on' in 'ext_params.py'.
    
    Parameters
    ----------
    :param saved_idx : the index into the set of extinction parameters saved in this function
        :type : int
    '''
    
    # use the default values found in 'ext_params.py'
    if saved_idx == 0:
        pass
    
    elif saved_idx == 1:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 2:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 3:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 4:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 5:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 6:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 7:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 8:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 9:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 10:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 11:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 12:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 13:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 14:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 15:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 16:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 17:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 18:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 19:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 20:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 21:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 22:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 23:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 24:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 25:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 26:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 27:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 28:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 29:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 30:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 31:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 32:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 33:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 34:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 35:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 36:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 37:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 38:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 39:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 40:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 41:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 42:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 43:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 44:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 45:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 46:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 47:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 48:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 49:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 50:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 51:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 52:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 53:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 54:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 55:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 56:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 57:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 58:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 59:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 60:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 61:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 62:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 63:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 64:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 65:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 66:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 67:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 68:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 69:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 70:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 71:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 72:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 73:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 74:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 75:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 76:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 77:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 78:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 79:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 80:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 81:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 82:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 83:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 84:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 85:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 86:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 87:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 88:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 89:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 90:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 91:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 92:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 93:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 94:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 95:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 96:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 97:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 98:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 99:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 100:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 101:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 102:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 103:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 104:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 105:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 106:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 107:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 108:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[0]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 109:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 110:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 111:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 112:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 113:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 114:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 115:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 116:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 117:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 118:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 119:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 120:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 121:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 122:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 123:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 124:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 125:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 126:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 127:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 128:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 129:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 130:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 131:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 132:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 133:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 134:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 135:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 136:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 137:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 138:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 139:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 140:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 141:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 142:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 143:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 144:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 145:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 146:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 147:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 148:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 149:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 150:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 151:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 152:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 153:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 154:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 155:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 156:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 157:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 158:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 159:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 160:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 161:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 162:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 163:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 164:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 165:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 166:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 167:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 168:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 169:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 170:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 171:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 172:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 173:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 174:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 175:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 176:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 177:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 178:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 179:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 180:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 181:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 182:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 183:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 184:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 185:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 186:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 187:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 188:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 189:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 190:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 191:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 192:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 193:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 194:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 195:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 196:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 197:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 198:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 199:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 200:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 201:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 202:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 203:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 204:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 205:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 206:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 207:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 208:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 209:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 210:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 211:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 212:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 213:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 214:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 215:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 216:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[1]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 217:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 218:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 219:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 220:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 221:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 222:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 223:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 224:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 225:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 226:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 227:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 228:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 229:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 230:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 231:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 232:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 233:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 234:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 235:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 236:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 237:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 238:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 239:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 240:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 241:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 242:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 243:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 244:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 245:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 246:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 247:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 248:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 249:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 250:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 251:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 252:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[0]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 253:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 254:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 255:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 256:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 257:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 258:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 259:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 260:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 261:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 262:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 263:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 264:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 265:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 266:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 267:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 268:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 269:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 270:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 271:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 272:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 273:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 274:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 275:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 276:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 277:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 278:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 279:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 280:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 281:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 282:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 283:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 284:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 285:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 286:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 287:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 288:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[1]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 289:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 290:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 291:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 292:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 293:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 294:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 295:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 296:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 297:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = True
        ep.prob_extinction_on = False
        ep.fit_extinction_on = False
        
        # interval extinction parameters
        ep.i_ext_interval = testing_i_ext_interval[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 298:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 299:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 300:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 301:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 302:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 303:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 304:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 305:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 306:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[0]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 307:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 308:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 309:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 310:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 311:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 312:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 313:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 314:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 315:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[1]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 316:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 317:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 318:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[0]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 319:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 320:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 321:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[1]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 322:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[0]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 323:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[1]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
    
    elif saved_idx == 324:
        # extinction parameters
        ep.extinction_on = True
        ep.ext_percent = testing_ext_percent[2]
        ep.ext_duration = testing_ext_duration[2]
        ep.ext_sort_order = ext_sort_order
        ep.ext_kills_least_fit = ext_kills_least_fit
        ep.ext_kills_constant = ext_kills_constant
        ep.ext_virulence = ext_virulence
        ep.ext_pop_dynamics_on = True
        ep.ext_pop_dyn_strength = ext_pop_dyn_strength
        ep.ext_end_extinction = True
        ep.ext_end_percent = ext_end_percent
        ep.ext_safeguard = ext_safeguard
        
        # elitism parameters
        ep.ext_elite_on = True
        ep.ext_elite_percent = ext_elite_percent
        
        # anti-elitism parameters
        ep.ext_anti_elite_on = True
        ep.ext_anti_elite_percent = ext_anti_elite_percent
        
        # extinction types
        ep.interval_extinction_on = False
        ep.prob_extinction_on = True
        ep.fit_extinction_on = False
        
        # probabilistic extinction parameters
        ep.p_ext_interval = testing_p_ext_interval[2]
        ep.p_ext_always_resets = p_ext_always_resets
        ep.p_ext_base_prob = p_ext_base_prob
        ep.p_ext_prob_inc = testing_p_ext_prob_inc[2]
        
        # repopulation parameters
        ep.repop_duration = testing_repop_duration[2]
        ep.repop_constant = False
        ep.repop_use_elites = True
        ep.repop_elite_percent = repop_elite_percent
        ep.repop_use_survivors = True
        ep.repop_survivor_percent = repop_survivor_percent
        
        # newborn parameters
        ep.newborn_params_on = True
        ep.newborn_expires = newborn_expires
        ep.newborn_protection_on = True
        
        # alternate parameters
        ep.alt_params_on = True
        
        # alternate selection parameters
        ep.alt_select_on = True
        ep.alt_tourn_size = alt_tourn_size
        
        # alternate crossover parameters
        ep.alt_xo_on = True
        ep.alt_xo_prob = alt_xo_prob
        
        # alternate mutation parameters
        ep.alt_mut_on = True
        ep.alt_mut_prob = alt_mut_prob
        ep.alt_mut_rate = alt_mut_rate
